# encoding: utf-8

"""
Utility functions for the `catcoocc` library.

The module mostly includes functions for input/output, as well as
pre-computations from raw data.
"""

# TODO: add function for smoothing of co-occurrences after collect_cooccs()
# TODO: add function for reading mushroom and pokemon data

# Import Python standard libraries
import csv
from collections import Counter
from itertools import chain, combinations, product

# Import 3rd party libraries
import numpy as np


def collect_alphabets(cooccs):
    """
    Return the `x` and `y` alphabets from a list of co-occurrences.

    :param list cooccs: The list of co-occurrence tuples.

    :return tuple alphabets: A tuple of two elements, the sorted list of
        symbols in series `x` and the sorted list of symbols in series `y`.
    """

    alphabet_x, alphabet_y = zip(*cooccs)

    return sorted(set(alphabet_x)), sorted(set(alphabet_y))


def collect_ngrams(seq, order, pad):
    """
    Function for yielding the ngrams of a sequence.

    The sequence is padded so that symbols at the extremities will be
    repeated the same number of times that other symbols. This operation also
    guarantees that sequences shorter than the order will be collected.

    :param list seq: The list of elements for ngram collection.
    :param int order: The ngram order.
    :param string pad: The padding symbol.

    :yield: The ngrams of the sequence.
    """

    seq = tuple(chain((pad,) * (order - 1), seq, (pad,) * (order - 1)))

    for ngram in zip(*[seq[i:] for i in range(order)]):
        yield ngram


def collect_cooccs(seqs, order=None, pad="#"):
    """
    Collects tuples of co-occurring elements in pairs of sequences.

    This function is used for collecting co-occurring elements in order to
    train or improve scorers. Co-occurrences can be collected either
    considering sequences in their entirety, or by aligned ngrams.

    When collecting ngrams, a padding symbol is used internally to guarantee
    that elements at the extremities of the sequences will be collected the
    same number of times that other elements. The padding symbol defaults to
    `"#"`, but given that the tuples where it occurs are automatically
    removed before returning, it is important that such symbol is not a
    part of either of the sequence alphabets. It is the user responsability
    to set an appropriate, non-conflicting pad symbol in case the default
    is used in the data.

    :param list seqs: A list of lists of sequence pairs.
    :param number order: The order of the ngrams to be collected, with `None`
        indicating the collection of the entire sequences (default: None).
    :param string pad: The symbol for internal padding, which must not
        conflict with symbols in the sequences (default: "#").

    :return list: A list of tuples of co-occurring elements.
    """

    # If an ngram order is specified, instead of looking for co-occurrences
    # across the full lenght of the sequences of a pair we need to take
    # overlapping ngrams in those (which must be aligned or, at least,
    # have the same length), creating a new set of sequences to get the
    # cooccurrences which is actually a set of ngram subsequences
    if not order:
        # No ngrams, just copy the sequence pairs, preserving the original
        _seqs = seqs[:]
    else:
        # For ngram collection, we need to guarantee that each in each
        # sequence pair both the sequences have the same length (assuming
        # they are aligned).
        len_check = [len(seq_a) == len(seq_b) for seq_a, seq_b in seqs]
        if not all(len_check):
            raise ValueError("Sequence pair of different lengths.")

        # Collect the ngrams for each sequence in each pair
        ngram_seqs = [
            [
                collect_ngrams(seq_a, order, pad),
                collect_ngrams(seq_b, order, pad),
            ]
            for seq_a, seq_b in seqs
        ]

        # Decompose the list of ngrams in `ngram_seqs`, building new
        # corresponding pairs from where the co-occurrences will be
        # collected. This is done with a non-expansive itertools operation,
        # with no need to cast a list (as the iterable will be consumed
        # into a list below).
        _seqs = chain(*[zip(*ngram_seq) for ngram_seq in ngram_seqs])

    # From the list of sequence pairs, chain a list with all the co-occurring
    # pairs from the product of each pair.
    coocc = chain(*[product(seq_a, seq_b) for seq_a, seq_b in _seqs])

    # Remove `(pad, pad)` entries that will result form ngram collection;
    # the operation is performed even in case of no `order` (i.e., collection
    # over the entire strings), as it allows us to consume `coocc` (at this
    # point, an iterable) into a list.
    coocc = [pair for pair in coocc if pad not in pair]

    return coocc


def collect_observations(cooccs):
    """
    Build a dictionary of observations for all possible correspondences.

    The counts of observations are derived from typical organization in
    a non-squared contingency table, using indexes 0, 1, and 2, where
    0 means that the identity is not considered (so that "00" would refer to
    all correspondence pairs), 1 that there is a match, and 2 that there is
    a mismatch. In more detail, given a pair of symbols `x` and `y`,
    the counts of co-occurrences will be:

    - obs["10"]: number of pairs matching `x` (with any `y`)
    - obs["20"]: number of pairs not matching `x` (with any `y`)
    - obs["01"]: number of pairs matching `y` (with any `x`)
    - obs["02"]: number of pairs not matching `y` (with any `x`)
    - obs["11"]: number of pairs matching `x` and matching `y`
    - obs["12"]: number of pairs matching `x` and not matching `y`
    - obs["21"]: number of pairs not matching `x` and matching `y`
    - obs["22"]: number of pairs not matching `x` and matching `Y`

    Note that obs["22"] is not the number of pairs where either `x` or `y`
    are mismatching, but where both necessarily mismatch.

    When doing x->y, a non-squared contingency table will be

        |            |  pair[1]  | pair[1]==y | pair[1]!=y |
        | pair[0]==x | obs["10"] | obs["11"]  | obs["12"]  |
        | pair[0]!=x | obs["20"] | obs["21"]  | obs["22"]  |

    :param list cooccs: A list of co-occurrence tuples.

    :return dict obs: A dictionary of observations per co-occurrence type.
    """

    # Collect observation counts; while this could be done in linear fashion,
    # especially with Python it takes too long for real data, so we need to
    # properly cache and pre-compute stuff at the expense of memory.
    #
    # "00": len(cooccs)
    # "10": len([p for p in cooccs if p[0] == a])
    # "20": len([p for p in cooccs if p[0] != a])
    # "01": len([p for p in cooccs if p[1] == b])
    # "02": len([p for p in cooccs if p[1] != b])
    # "11": cooccs.count((a, b))
    # "22": len([p for p in cooccs if not any([p[0] == a, p[1] == b])])
    # "12": len([p for p in cooccs if p[0] == a and p[1] != b])
    # "21": len([p for p in cooccs if p[0] != a and p[1] == b])

    # Cache the number of co-occurrences, which is obs_00 for all pairs
    obs_00 = len(cooccs)

    # Obs['11'] is cached with a collections.Counter, so that we rely on
    # the standard library
    obs_11 = Counter(cooccs)

    # Extract the `x` and `y` symbols, so they can be counted as well;
    # the alphabets can be built from their respective sets. We don't
    # use collect_alphabet() as we need the symbols as well, in order to
    # build `x_counter` and `y_counter`
    symbols_x, symbols_y = zip(*cooccs)
    x_counter, y_counter = Counter(symbols_x), Counter(symbols_y)
    alphabet_x = set(symbols_x)
    alphabet_y = set(symbols_y)

    # Build with a dictionary comprehension
    obs = {
        (x, y): {
            "00": obs_00,
            "11": obs_11[(x, y)],
            "10": x_counter[x],
            "20": obs_00 - x_counter[x],
            "01": y_counter[y],
            "02": obs_00 - y_counter[y],
            "12": x_counter[x] - obs_11[(x, y)],
            "21": y_counter[y] - obs_11[(x, y)],
            "22": obs_00 + obs_11[(x, y)] - x_counter[x] - y_counter[y],
        }
        for x, y in product(alphabet_x, alphabet_y)
    }

    return obs


def build_ct(observ, square):
    """
    Build a contingency table from a dictionary of observations.

    The contingency table can be either square (2x2) or not (3x2). Non-squared
    contingency tables include co-occurrences where neither the `x` or `y`
    under investigation occur.

    :param dict observ: A dictionary of observations, as provided by
        utils.collect_observations()
    :param bool square: Whether to return a square (2x2) or non-square (3x2)
            contingency table.
    """

    # Build the contingency tables as np.arrays(), as they will necessary
    # for the functions consuming them and as it allows to use np methods
    # directly (such as .sum())
    if square:
        cont_table = np.array(
            [[observ["11"], observ["12"]], [observ["21"], observ["22"]]]
        )
    else:
        cont_table = np.array(
            [
                [observ["10"], observ["11"], observ["12"]],
                [observ["20"], observ["21"], observ["22"]],
            ]
        )

    return cont_table


def read_sequences(filename, cols=None, col_delim="\t", elem_delim=" "):
    """
    Reads parallel sequences, returning them as a list of lists.

    Datafiles should have one sequence pair per row, with pairs separated
    by `col_delim` and elements delimited by `elem_delim`, as in:

    ```
    Orthography                 Segments
    E X C L A M A T I O N       ɛ k s k l ʌ m eɪ ʃ ʌ n
    C L O S E Q U O T E         k l oʊ z k w oʊ t
    D O U B L E Q U O T E       d ʌ b ʌ l k w oʊ t
    E N D O F Q U O T E         ɛ n d ʌ v k w oʊ t
    E N D Q U O T E             ɛ n d k w oʊ t
    I N Q U O T E S             ɪ n k w oʊ t s
    Q U O T E                   k w oʊ t
    U N Q U O T E               ʌ n k w oʊ t
    H A S H M A R K             h æ m ɑ ɹ k
    ```

    Rows without complete data are not included. The function allows to
    collect data from multiple columns, but it is inteded to pairwise
    comparison; if no column names are provided, as per default,
    the reading will skip the first row (assumed to be header).

    :param str filename: Path to the file to be read.
    :param cols list: List of column names to be collected (default: None)
    :param str col_delim: String used as field delimiter (default: `"\t"`).
    :param str elem_delim: String used as element delimiter (default: `" "`).
    """

    # Column names must always be specified, otherwise we will skip over
    # the header (thus assuming that there are only two columns); the logic
    # is different
    data = []
    with open(filename) as handler:
        if not cols:
            skip_header = True
            for line in handler.readlines():
                if skip_header:
                    skip_header = False
                    continue

                data.append(
                    [
                        column.split(elem_delim)
                        for column in line.strip().split(col_delim)
                    ]
                )
        else:
            reader = csv.DictReader(handler, delimiter=col_delim)
            data = [
                [
                    row[col_name].split(elem_delim)
                    for col_name in cols
                    if row[col_name]
                ]
                for row in reader
            ]

    # Remove incomplete rows
    # NOTE: Checking length is not really effective, but will allow easier
    # expansion in the future
    data = [row for row in data if len(row) == 2]

    return data


# TODO: assuming one taxon per col and locations per row, allow to switch
def read_pa_matrix(filename, delimiter="\t"):
    """
    Reads a presence-absence matrix, returning as equivalent sequences.

    The location must be specified under column name `ID`.

    :param str filename: Path to the file to be read.
    :param str delimiter: String used as field delimiter (default: `"\t"`).
    """

    # We read the matrix and filter each row (location), keeping only
    # the observed taxa
    matrix = {}
    with open(filename) as csvfile:
        reader = csv.DictReader(csvfile, delimiter=delimiter)
        for row in reader:
            location = row.pop("ID")
            matrix[location] = sorted(
                [taxon for taxon, observed in row.items() if observed == "1"]
            )

    # Make the corresponding sequence from the combinations
    obs_combinations = list(
        chain.from_iterable(
            [combinations(observed, 2) for location, observed in matrix.items()]
        )
    )

    return obs_combinations



# encoding: utf-8

# Note: includes code modified from
# - https://stackoverflow.com/questions/46498455/categorical-features-correlation/46498792
# - https://towardsdatascience.com/the-search-for-categorical-correlation-a1cf7f1888c9

"""
Defines the various scorers for categorical co-occurrence analysis.
"""

# TODO: have a function for checking if we have enough data for a chi2, combined with smoothing
# TODO: improve the speed of theil's u computation
# TODO: extend scipy options in chi2_contingency
# TODO: investigate if it is worth reusing chi2 statistics for cramer
# TODO: include a logarithmic scaler (instead of percentile one)
# TODO: allow to combine scalers? log within range?

# Import Python standard libraries
from collections import Counter
from itertools import chain, product
import math

# Import 3rd party libraries
import numpy as np
import scipy.stats as ss

# import local libraries
from . import utils


def conditional_entropy(x_symbols, y_symbols):
    """
    Computes the entropy of `x` given `y`.

    :param list x_symbols: A list of all observed `x` symbols.
    :param list y_symbols: A list of all observed `y` symbols.

    :return float entropy: The conditional entropy of `x` given `y`.
    """

    # Cache counters; while the xy_counter is already computed in other
    # parts, particularly in the scorers, it is worth repeating the code
    # here to have a more general function.
    y_counter = Counter(y_symbols)
    xy_counter = Counter(list(zip(x_symbols, y_symbols)))
    population = sum(y_counter.values())

    # Compute the entropy and return
    entropy = 0
    for xy_pair, xy_count in xy_counter.items():
        p_xy = xy_count / population
        p_y = y_counter[xy_pair[1]] / population
        entropy += p_xy * math.log(p_y / p_xy)

    return entropy


def compute_cramers_v(cont_table):
    """
    Compute Cramer's V from a contingency table.

    :param np.array cont_table: The contingency table for computation.

    :return float v: The Cramér's V measure for the given contingency table.
    """

    # Cache the shape and sum of the contingency table
    rows, cols = cont_table.shape
    population = cont_table.sum()

    # Compute chi2 and phi2
    chi2 = ss.chi2_contingency(cont_table)[0]
    phi2 = chi2 / population

    # Compute the correlations for Cramér's V
    phi2corr = max(0, phi2 - ((cols - 1) * (rows - 1)) / (population - 1))
    rcorr = rows - ((rows - 1) ** 2) / (population - 1)
    kcorr = cols - ((cols - 1) ** 2) / (population - 1)

    # Compute Cramér's from the correlations
    return np.sqrt(phi2corr / min((kcorr - 1), (rcorr - 1)))


def compute_pmi(p_x, p_y, p_xy, normalized, limit=1e-6):
    """
    Compute the Pointwise Mutual Information.

    :param float p_x: The probability of `x`, p(x).
    :param float p_y: The probability of `y`, p(y).
    :param float p_xy: The probability of `xy`, p(xy).
    :param bool normalized: Whether to return the normalized Pointwise
        Mutual Information in range [-1, 1].
    :param float limit: The value to use for computation when `p_xy` is 0.0
        (i.e., non observed), as the limit of PMI to zero (default: 1e-6).

    :return float pmi: The Pointwise Mutual Information.
    """

    if p_xy == 0.0:
        pmi = math.log(limit / (p_x * p_y))
        if normalized:
            pmi = pmi / -math.log(limit)
    else:
        pmi = math.log(p_xy / (p_x * p_y))
        if normalized:
            pmi = pmi / -math.log(p_xy)

    return pmi


def compute_theil_u(x_symbols, y_symbols):
    """
    Compute the uncertainty coefficient Theil's U.

    :param list x_symbols: The list of observed symbols in series `x`.
    :param list y_symbols: The list of observed symbols in series `y`.

    :return float theil_u : The uncertainty coefficient given `x` and `y`.
    """

    # Compute the conditional entropy of `x` given `y`
    h_xy = conditional_entropy(x_symbols, y_symbols)

    # Cache a counter for the symbols of the `x` series and its length
    x_counter = Counter(x_symbols)
    population = len(x_symbols)

    # Compute the probability for each symbol and, from there, the entropy
    p_x = list(map(lambda n: n / population, x_counter.values()))
    h_x = ss.entropy(p_x)

    # If the entropy is zero (all items alike), the uncertainty coeffiecient
    # is by definition 1.0; otherwise, compute it in relation to the
    # conditional entropy
    if h_x == 0.0:
        theil_u = 1.0
    else:
        theil_u = (h_x - h_xy) / h_x

    return theil_u


# TODO: allow independent scaling over `x` and independent over `y` (currently doing all)
# TODO: allow scaling withing percentile borders
# TODO: see if we can vectorize numpy operations (now on dicts)
def scale_scorer(scorer, method="minmax", nrange=(0, 1)):
    """
    Scale a scorer.

    The function returns a scaled version of a scorer considering all
    the assymmetric scorers (i.e., both `x` given `y` and `y` given `x`).
    Implemented scoring methods are "minmax" (by default on range [0, 1],
    which can be modified by the `nrange` parameter) and "mean".

    :param dict scorer: A scoring dictionary.
    :param str method: The scoring method to be used, either `"minmax"` or
        `"mean"` or `"stdev"` (default: `"minmax"`).
    :param tuple nrange: A tuple with the scaling range, to be used when
        applicable (default: (0, 1)).

    :return dict scaled_scorer: A scaled version of the scorer.
    """

    # Extract scores as a list, combining `xy` and `yx`
    scores = list(chain.from_iterable(scorer.values()))

    # normalizatoin is performed over xy and yx together
    if method == "minmax":
        # Set normalization range
        range_low, range_high = nrange

        # Cache values for speeding computation
        min_score = min(scores)
        max_score = max(scores)
        score_diff = max_score - min_score
        range_diff = range_high - range_low

        scaled_scorer = {
            pair: (
                range_low
                + (((value[0] - min_score) * range_diff) / score_diff),
                range_low
                + (((value[1] - min_score) * range_diff) / score_diff),
            )
            for pair, value in scorer.items()
        }

    elif method == "mean":
        # Cache values for speeding computation
        mean = np.mean(scores)
        score_diff = max(scores) - min(scores)

        scaled_scorer = {
            pair: (
                (value[0] - mean) / score_diff,
                (value[1] - mean) / score_diff,
            )
            for pair, value in scorer.items()
        }
    elif method == "stdev":
        mean = np.mean(scores)
        stdev = np.std(scores)

        scaled_scorer = {
            pair: ((value[0] - mean) / stdev, (value[1] - mean) / stdev)
            for pair, value in scorer.items()
        }
    else:
        raise ValueError("Unknown scaling method.")

    return scaled_scorer


def scorer2matrices(scorer):
    """
    Return the asymmetric matrices implied by a scorer and their alphabets.

    :param dict scorer: A scoring dictionary.

    :return np.array xy: A scoring matrix of `y` given `x`.
    :return np.array yx: A scoring matrix of `x` given `y`.
    :return list alphabet_x: The alphabet for matrix `x`.
    :return list alphabet_y: The alphabet for matrix `y`.
    """

    alphabet_x, alphabet_y = utils.collect_alphabets(scorer)

    xy = np.array(
        [np.array([scorer[(x, y)][0] for x in alphabet_x]) for y in alphabet_y]
    )
    yx = np.array(
        [np.array([scorer[(x, y)][1] for y in alphabet_y]) for x in alphabet_x]
    )

    return xy, yx, alphabet_x, alphabet_y


class CatScorer:
    """
    Class for computing catagorical co-occurrence scores.
    """

    def __init__(self, cooccs):
        """
        Initialization function.
        """

        # Store co-occs, observations, symbols and alphabets
        self.cooccs = cooccs
        self.obs = utils.collect_observations(cooccs)

        # Obtain the alphabets from the co-occurrence pairs
        self.alphabet_x, self.alphabet_y = utils.collect_alphabets(self.cooccs)

        # Initialize the square and non-square contingency table as None
        self._square_ct = None
        self._nonsquare_ct = None

        # Initialize the scorers as None, in a lazy fashion
        self._mle = None
        self._pmi = None
        self._npmi = None
        self._chi2 = None
        self._chi2_nonsquare = None
        self._cramersv = None
        self._cramersv_nonsquare = None
        self._fisher = None
        self._theil_u = None
        self._cond_entropy = None
        self._tresoldi = None

    def _compute_contingency_table(self, square):
        """
        Internal method for computing and caching contingency tables.

        :param bool square: Whether to compute a square (2x2) or non-square
            (3x2) contingency table.
        """

        if square and not self._square_ct:
            self._square_ct = {
                pair: utils.build_ct(self.obs[pair], True) for pair in self.obs
            }

        if not square and not self._nonsquare_ct:
            self._nonsquare_ct = {
                pair: utils.build_ct(self.obs[pair], False) for pair in self.obs
            }

    def mle(self):
        """
        Return an MLE scorer, computing it if necessary.
        """

        # Compute the scorer, if necessary
        if not self._mle:
            self._mle = {
                pair: (
                    self.obs[pair]["11"] / self.obs[pair]["10"],
                    self.obs[pair]["11"] / self.obs[pair]["01"],
                )
                for pair in product(self.alphabet_x, self.alphabet_y)
            }

        return self._mle

    def pmi(self, normalized=False):
        """
        Return a PMI scorer.

        :param bool normalize: Whether to return a normalized PMI or
            not (default: False)
        """

        # Compute the non-normalized scorer, if necessary
        if not normalized and not self._pmi:
            # Use (1/n)^2 as the limit for unobserved pairs
            limit = (1 / len(self.cooccs)) ** 2

            self._pmi = {}
            for pair in self.obs:
                p_xy = self.obs[pair]["11"] / self.obs[pair]["00"]
                p_x = self.obs[pair]["10"] / self.obs[pair]["00"]
                p_y = self.obs[pair]["01"] / self.obs[pair]["00"]

                self._pmi[pair] = (
                    compute_pmi(p_x, p_y, p_xy, False, limit),
                    compute_pmi(p_y, p_x, p_xy, False, limit),
                )

        # Compute the normalized scorer, if necessary
        if normalized and not self._npmi:
            # Use (1/n)^2 as the limit for unobserved pairs
            limit = (1 / len(self.cooccs)) ** 2

            self._npmi = {}
            for pair in self.obs:
                p_xy = self.obs[pair]["11"] / self.obs[pair]["00"]
                p_x = self.obs[pair]["10"] / self.obs[pair]["00"]
                p_y = self.obs[pair]["01"] / self.obs[pair]["00"]

                self._npmi[pair] = (
                    compute_pmi(p_x, p_y, p_xy, True, limit),
                    compute_pmi(p_y, p_x, p_xy, True, limit),
                )

        # Select the scorer to return (this allows easier refactoring later)
        if not normalized:
            ret = self._pmi
        else:
            ret = self._npmi

        return ret

    def chi2(self, square_ct=True):
        """
        Return a Chi2 scorer.

        :param bool square_ct: Whether to return the score over a squared or
            non squared contingency table (default: True)
        """

        # Compute the scorer over a square contingency table, if necessary
        if square_ct and not self._chi2:
            # Compute the square contingency table, if necessary
            self._compute_contingency_table(True)

            # Build the scorer
            self._chi2 = {}
            for pair in self.obs:
                chi2 = ss.chi2_contingency(self._square_ct[pair])[0]
                self._chi2[pair] = (chi2, chi2)

        # Compute the scorer over a non-square contingency table, if necessary
        if square_ct and not self._chi2_nonsquare:
            # Compute the non square contingency table, if necessary
            self._compute_contingency_table(False)

            # Build the scorer
            self._chi2_nonsquare = {}
            for pair in self.obs:
                chi2 = ss.chi2_contingency(self._nonsquare_ct[pair])[0]
                self._chi2_nonsquare[pair] = (chi2, chi2)

        # Select the scorer to return (this allows easier refactoring later)
        if square_ct:
            ret = self._chi2
        else:
            ret = self._chi2_nonsquare

        return ret

    def cramers_v(self, square_ct=True):
        """
        Return a Cramér's V scorer.

        :param bool square_ct: Whether to return the score over a squared or
            non squared contingency table (default: True)
        """

        # Compute the scorer over a square contingency table, if necessary
        if square_ct and not self._cramersv:
            # Compute the square contingency table, if necessary
            self._compute_contingency_table(True)

            # Build the scorer
            self._cramersv = {}
            for pair in self.obs:
                cramersv = compute_cramers_v(self._square_ct[pair])
                self._cramersv[pair] = (cramersv, cramersv)

        # Compute the scorer over a non-square contingency table, if necessary
        if square_ct and not self._cramersv_nonsquare:
            # Compute the non square contingency table, if necessary
            self._compute_contingency_table(False)

            # Build the scorer
            self._cramersv_nonsquare = {}
            for pair in self.obs:
                cramersv = compute_cramers_v(self._nonsquare_ct[pair])
                self._cramersv_nonsquare[pair] = (cramersv, cramersv)

        # Select the scorer to return (this allows easier refactoring later)
        if square_ct:
            ret = self._cramersv
        else:
            ret = self._cramersv_nonsquare

        return ret

    def fisher(self):
        """
        Return a Fisher Exact Odds Ratio scorer.

        Please note that in scipy's implementation the calculated odds ratio
        is different from the one found in R. While the latter returns the
        "conditional Maximum Likelihood Estimate", this implementation computes
        the more common "unconditional Maximum Likelihood Estimate", which
        is known to be very slow for contingency tables with large numbers.
        If the computation is too slow, the similar but inexact chi-square
        test the the chi2 contingency scorer can be used.
        """

        # Compute the square contingency table, if necessary
        self._compute_contingency_table(True)

        # Compute the scorer, if necessary
        if not self._fisher:
            self._fisher = {}
            for pair in self.obs:
                fisher = ss.fisher_exact(self._square_ct[pair])[0]
                self._fisher[pair] = (fisher, fisher)

        return self._fisher

    def theil_u(self):
        """
        Return a Theil's U uncertainty scorer.
        """

        # Compute theil u, if necessary; the code uses two nested loops
        # instead of a product(x, y) to gain some speed
        if not self._theil_u:
            self._theil_u = {}

            for x in self.alphabet_x:
                for y in self.alphabet_y:
                    subset = [
                        pair
                        for pair in self.cooccs
                        if pair[0] == x or pair[1] == y
                    ]
                    X = [pair[0] for pair in subset]
                    Y = [pair[1] for pair in subset]

                    # run theil's
                    self._theil_u[(x, y)] = (
                        compute_theil_u(Y, X),
                        compute_theil_u(X, Y),
                    )

        return self._theil_u

    def cond_entropy(self):
        """
        Return a corrected conditional entropy scorer.
        """

        if not self._cond_entropy:
            self._cond_entropy = {}

            for x in self.alphabet_x:
                for y in self.alphabet_y:
                    subset = [
                        pair
                        for pair in self.cooccs
                        if pair[0] == x or pair[1] == y
                    ]
                    X = [pair[0] for pair in subset]
                    Y = [pair[1] for pair in subset]

                    # run theil's
                    self._cond_entropy[(x, y)] = (
                        conditional_entropy(Y, X),
                        conditional_entropy(X, Y),
                    )

        return self._cond_entropy

    def tresoldi(self):
        """
        Return a `tresoldi` asymmetric uncertainty scorer.

        This is our intended scorer for alignments.
        """

        # Build the scorer, if necessary
        if not self._tresoldi:
            # Obtain the MLE and PMI scorers for all pairs (which will
            # trigger their computation, if necessary)
            mle = self.mle()
            pmi = self.pmi()

            # Build the new scorer
            self._tresoldi = {}
            for pair in self.obs:
                if pmi[pair][0] < 0:
                    xy = -((-pmi[pair][0]) ** (1 - mle[pair][0]))
                else:
                    xy = pmi[pair][0] ** (1 - mle[pair][0])

                if pmi[pair][1] < 0:
                    yx = -((-pmi[pair][1]) ** (1 - mle[pair][1]))
                else:
                    yx = pmi[pair][1] ** (1 - mle[pair][1])

                self._tresoldi[pair] = (xy, yx)

        return self._tresoldi



# __init__.py

# Informatin on the package
__version__ = "0.2"
__author__ = "Tiago Tresoldi"
__email__ = "tresoldi@shh.mpg.de"

# Build the namespace
from catcoocc.utils import *
from catcoocc import scorer

# Resource dir (mostly for tests)
from pathlib import Path

RESOURCE_DIR = Path(__file__).parent.parent / "resources"



# -*- coding: utf-8 -*-
"""Demo.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/github/tresoldi/catcoocc/blob/master/docs/Demo.ipynb

# Categorical Co-occurrences

`catcoocc` is a library for facilitating the analysis of categorical co-occurrence by means of symmetric matrices, asymmetric matrices, and association rules. It is intended for usage on three main types of data: collections of pairwise sequences (both aligned and not), categorical fields in databases, and matrices of presence/absence (such as for investigation of species co-occurrence in biology). More than statistical significance, interested in effect size and strength of association.

For the purposes of the library, the following definitions are used:

- a co-occurrence is an observation of two categorical elements from two different series
- a population is a collection of co-occurrences which can come from the product of elements observed in two sequences (or, if the sequences are aligned, by n-gram windows), by the properties of an individual in a population, or by the product of presences
- a scorer is a collection of association measures between all the possible products of values in alphabet_x and alphabet_y (including those not observed in the data)
- scorers can be computed according to different methods (such as MLE...), either on raw counts of observed data or on smoothed methods; likewise, the output of scorers can be scaled in different ways
- scorers can be symmetric (such as those coming from chi2 methods) or assymetric (such a Theil U), and can be either restricted to positive values or include negative values, and can be bound or not to a range
"""

# This cell contains the system commands for locally installing the package;
# please remember that in most cases, especially in the playground,
# Google CoLab requires the runtime to be restarted (CTRL+M) in order for it
# to pick the new package (thus: run this cell, restart the runtime, and
# run other cells)
#!rm -rf catcoocc
#!git clone https://github.com/tresoldi/catcoocc.git
#!pip3 install --force-reinstall --no-deps -e catcoocc/

#from IPython.display import HTML, display

# Import 3rd party libraries
from matplotlib import pyplot as plt
import pandas as pd
import numpy as np
import seaborn as sns
import tabulate

# Import own libraries
import catcoocc
from catcoocc.scorer import CatScorer

def plot_scorer(scorer, alpha_x, alpha_y, title=None, font_scale=2, figsize=(25, 25)):
    if not title:
        title = ""

    matrix = pd.DataFrame(scorer, index=alpha_y, columns=alpha_x)

    sns.set(font_scale=font_scale)
    plt.figure(figsize=figsize)

    ax = plt.subplot(111)
    sns.heatmap(matrix, annot=True, fmt='.2f', linewidths=.5, center=0, ax=ax).set_title(title, fontsize=font_scale*figsize[0])

"""Let's start exploring using a subset of the mushroom dataset.

| class     | cap-shape |
|-----------|-----------|
| poisonous | convex    |
| edible    | convex    |
| edible    | bell      |
| poisonous | convex    |
| edible    | convex    |
| edible    | convex    |
| edible    | bell      |
| edible    | bell      |
| poisonous | convex    |
| edible    | bell      |
| edible    | convex    |
| edible    | convex    |
| edible    | bell      |
| poisonous | convex    |
| edible    | convex    |
| edible    | sunken    |
| edible    | flat      |
| poisonous | convex    |
| poisonous | convex    |
| poisonous | convex    |

Mushrooms can be either edible or poisonous, and the cap shapre can be of four types, convex, bell, flat, or sunken. In this subset, flat and sunken are underrepresented, with a single case, both being edibles, but of course it is not safe to eat it?
"""

mushroom_data = catcoocc.read_sequences("docs/mushroom-small.tsv")
mushroom_cooccs = catcoocc.collect_cooccs(mushroom_data)
scorer = catcoocc.scorer.CatScorer(mushroom_cooccs)

mle = scorer.mle()
pmi = scorer.pmi()
npmi = scorer.pmi(True)
chi2 = scorer.chi2()
chi2_ns = scorer.chi2(False)
cramersv = scorer.cramers_v()
cramersv_ns = scorer.cramers_v(False)
fisher = scorer.fisher()
theil_u = scorer.theil_u()
catcoocc_i = scorer.catcoocc_i()
catcoocc_ii = scorer.catcoocc_ii()

headers = [
    'pair',
    'mle_0',        'mle_1',
    'pmi_0',        'pmi_1',
    'npmi_0',       'npmi_1',
    'chi2_0',       'chi2_1',
    'chi2ns_0',     'chi2ns_1',
    'cremersv_0',   'cremersv_1',
    'cremersvns_0', 'cremersvns_1',
    'fisher_0',     'fisher_1',
    'theilu_0',     'theilu_1',
    'catcoocci_0',  'catcoocci_1',
    'catcooccii_0', 'catcooccii_1',
]

table = []
for pair in sorted(scorer.obs):
    buf = [
        pair,
        "%0.4f" % mle[pair][0],         "%0.4f" % mle[pair][1],
        "%0.4f" % pmi[pair][0],         "%0.4f" % pmi[pair][1],
        "%0.4f" % npmi[pair][0],        "%0.4f" % npmi[pair][1],
        "%0.4f" % chi2[pair][0],        "%0.4f" % chi2[pair][1],
        "%0.4f" % chi2_ns[pair][0],     "%0.4f" % chi2_ns[pair][1],
        "%0.4f" % cramersv[pair][0],    "%0.4f" % cramersv[pair][1],
        "%0.4f" % cramersv_ns[pair][0], "%0.4f" % cramersv_ns[pair][1],
        "%0.4f" % fisher[pair][0],      "%0.4f" % fisher[pair][1],
        "%0.4f" % theil_u[pair][0],     "%0.4f" % theil_u[pair][1],
        "%0.4f" % catcoocc_i[pair][0],  "%0.4f" % catcoocc_i[pair][1],
        "%0.4f" % catcoocc_ii[pair][0], "%0.4f" % catcoocc_ii[pair][1],
    ]
    table.append(buf)

    
print(tabulate.tabulate(table, headers=headers, tablefmt='markdown'))

"""and now more"""

xy, yx, alpha_x, alpha_y = catcoocc.scorer.scorer2matrices(catcoocc_ii)
plot_scorer(xy, alpha_x, alpha_y, "x->y", 2, (15, 15))

"""and even more"""

plot_scorer(yx, alpha_y, alpha_x, "y->x", 2, (15, 15))



import pathlib
from setuptools import setup

# The directory containing this file
LOCAL_PATH = pathlib.Path(__file__).parent

# The text of the README file
README_FILE = (LOCAL_PATH / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="catcoocc",
    version="0.2",
    description="Methods for symmetric and asymmetric analysis of categorical co-occurrences",
    long_description=README_FILE,
    long_description_content_type="text/markdown",
    url="https://github.com/tresoldi/catcoocc",
    author="Tiago Tresoldi",
    author_email="tresoldi@shh.mpg.de",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
    packages=["catcoocc", "docs", "resources"],
    keywords=["co-occurrence", "cooccurrence", "categorical variables", "mutual information", "scorer"],
    include_package_data=True,
    install_requires=[
        "numpy",
        "scipy",
        "tabulate",
    ],
    entry_points={
        "console_scripts": [
            "catcoocc=catcoocc.__main__:main",
        ]
    },
    test_suite='tests',
    tests_require=[],
    zip_safe=False,
)

